/*************************************************************
*  This file is part of the Surface Evolver source code.     *
*  Programmer:  Ken Brakke, brakke@susqu.edu                 *
*************************************************************/


/**********************************************************
*
*  File: query.c
*
*  Contents:  query interpreter for Surface Evolver
*/

#include "include.h"
#include "ytab.h"

int commandverb;

/***********************************************************************
  History mechanism.
  Command texts are stored in linear array history_space[], separated by
  nulls, most recent first.  history_offsets[] contains indices of starts
  of commands, -1 for no command.
  history_number is permanent counter of commands.
  history_count is number of currently saved commands.
**************************************************************************/


/**************************************************************************
*  
* function: new_history()
*
* purpose:  Appends command to history list, queue fashion.
*/

void new_history ARGS1((text),
char *text)
{ size_t length;
  int k;

  if ( history_space == NULL )
  { history_space = my_list_calloc(HISTORYSPACE,1,ETERNAL_BLOCK);
    for ( k = 0 ; k < MAXHISTORY ; k++ ) history_offsets[k] = -1; /* empty */
    history_count = 0;
  }
  length = strlen(text)+1;
  if ( length >= HISTORYSPACE-5 ) 
  { kb_error(1585,"Command too long to save in history list.\n",WARNING);
    return;
  }
  kb_memmove(history_space+length,history_space,HISTORYSPACE-length);
  for ( k = history_count-1 ; k >= 0 ; k-- )
  { /* adjust pointers in history list */
    history_offsets[k+1] = history_offsets[k] + (int)length;
    if ( history_offsets[k+1] >= HISTORYSPACE )
       { history_offsets[k+1] = -1; history_count--; }
  }
  strcpy(history_space,text);
  history_offsets[0] = 0;
  if ( history_count < MAXHISTORY-1 ) history_count++;
  history_number++;
}

/**************************************************************************
*  
* function: old_history()
*
* purpose:  Checks commands beginning with ! for history list match.
* return:   retval from command()
*/

int old_history ARGS1((text),
char *text)
{ char *h;
  int k;
  int hnum;
  int  retval = 0;

  if ( history_space == NULL )
    { history_space = my_list_calloc(HISTORYSPACE,1,ETERNAL_BLOCK);
      for ( k = 0 ; k < MAXHISTORY ; k++ ) history_offsets[k] = -1; /* empty */
      history_count = 0;
    }

  if (text[1] == '!' ) 
    {
      if ( history_count > 0 ) 
      { k = 0; goto do_old;
      }
      else { kb_error(1586,"History list is empty.\n",WARNING); return retval; }

    }
  else if ( isdigit(text[1]) )
     { hnum = atoi(text+1);
       k = history_number - hnum;
       if ( (k >= 0) && (k < history_count) )
          goto do_old;
       else kb_error(1587,"Command not in history list.\n",RECOVERABLE);
     }
  else for ( k = 0 ; k < history_count ; k++ )
     { h = history_space+history_offsets[k]; 
       if ( strlen(h) < strlen(text+1) ) continue; /* prevent compare overrun */
       if ( strncmp(h,text+1,strlen(text+1)) == 0 )
          goto do_old;
     }
  kb_error(1588,"Command not found in history.\n",RECOVERABLE);


do_old:
  h = history_space + history_offsets[k];
  outstring(h);outstring("\n");
  retval = command(h,ADD_TO_HISTORY);  /* reinserts fulltext copy in history list */
  return retval;
}

/**************************************************************************
*  
* function: catfulltext()
*
* purpose:  catenates line to full text of command
*/

void catfulltext ARGS1((stuff),
char *stuff)
{

  if ( fulltextsize > MAXCMDSIZE-3 ) return; /* too long */
  if ( fulltextsize > 0 ) 
     { strcat(fulltext,"  "); /* indentation */
        fulltextsize += 2;
     }
  strncpy(fulltext+fulltextsize,stuff,MAXCMDSIZE-fulltextsize);
  fulltextsize = strlen(fulltext);
  if ( fulltext[fulltextsize-1] != '\n' ) fulltext[fulltextsize++] = '\n';
  fulltext[fulltextsize] = '\0';

}

/**************************************************************************
*  
* function: command()
*
* purpose:  Executes a text command string.
*
* return:  END_COMMANDS if current command interpreter loop is to end.
*          0 else
*/

int command ARGS2((text,mode),
char *text,
int mode) /* NO_HISTORY or ADD_TO_HISTORY */
{
  struct expnode qnode;  /* for query expression */
  int retval = 0;
  int old_scope;
  struct treenode *old_list;
  int old_listtop, old_listmax;
  int old_brace_depth = brace_depth;

  memset(&qnode,0,sizeof(struct expnode));
  sprintf(qnode.name,"command");
 
  old_listmax = listmax;
  old_list = list; old_listtop = listtop; /* for nested parsing */
  listmax = 30;
  list = (struct treenode *)mycalloc(listmax,sizeof(struct treenode));
  listtop = 1;
  loopdepth = 0;
  parse_error_flag = 0;
  lists_flag = 0;
  cmdptr = text;
  old_scope = begin_scope();


#ifdef __cplusplus
  try
  {
#else
  if ( setjmp(cmdbuf) )
  { set_scope(old_scope);  
    cmdptr = NULL;  
    myfree((char *)list); 
    list = NULL;
    memset(local_scope_bases,0,sizeof(local_scope_bases));
	local_nest_depth = 0;
    return 0; 
  }
#endif

  verb_flag = 1; /* tell lex we need a verb first */
  old_global_count = web.global_count; /* for error recovery */
  old_perm_global_count = web.perm_global_count; /* for error recovery */
  /* local_nest_depth = 0; */
  init_local_scope(0);
  fulltextsize = 0; 
  catfulltext(text);

  retval = yybegin();  /* 0 for accept, 1 for error */
  qnode.locals = localbase;
  if ( localbase )
    qnode.locals->flags |= LL_IN_USE;
  exit_local_scope();

  cmdptr = NULL;

  if ( (retval == 1) || (parse_error_flag) )
  { myfree((char *)list); 
    if ( list == permlist ) permlist = NULL;
    list = NULL; set_scope(old_scope); 
    return 0;  
  }

  qnode.start=list;
  qnode.root = list + listtop - 1;
  if ( qnode.root->type == 0 ) /* empty command */
  { retval = 0;
    goto command_exit;
  }
  if ( qnode.root->type != CMDLIST_ )  /* expression or something */
     kb_error(1589,"Illegal command.\n",COMMAND_ERROR);
  

  list[0] = list[listtop-1];  /* root also in first spot */
  if ( list[0].left ) list[0].left += listtop - 1;
  if ( list[0].right ) list[0].right += listtop - 1;

  /* put DONE marker after root */
  list[listtop++].type = FINISHED;

  /* figure stack usage */
  stack_usage(&qnode);

  /* set element loop flag bit in nodes */
  if ( !function_kludge_flag )
    mark_element_loops(&list[0],0);

  if ( logfd )
  fprintf(logfd,"%s\n",text);

  /* add to history list */
  if ( mode == ADD_TO_HISTORY ) 
    new_history(fulltext);

  /* initialize current element to NULLID */
  aggregate_depth = 0; calc_quant_flag = 0;
  exit_flag = 0;
  eval(&qnode,NULL,NULLID,NULL);
  retval = exit_flag ? END_COMMANDS : 1;
  exit_flag = 0;

command_exit:
  qnode.flag = USERCOPY; /* so free_expr will work */
  free_expr(&qnode); 
  list = NULL;
  set_scope(old_scope);
  list = old_list; listtop = old_listtop; /* for nested parsing */
  listmax = old_listmax;
  brace_depth = old_brace_depth;
  return retval;

#ifdef __cplusplus
  }
  catch(cmd_excep c)
  { set_scope(old_scope);  
    cmdptr = NULL;  
    myfree((char *)list); 
    list = NULL;
    memset(local_scope_bases,0,sizeof(local_scope_bases));
	local_nest_depth = 0;
    return 0; 
  }
#endif
}

/*************************************************************************
*
* function: stack_usage()
*
* purpose: figure maximum stack usage of expression or command.
*
*/

void stack_usage ARGS1((ex ),
struct expnode *ex)
{ int usage = 0;
  struct treenode *node ;
#ifdef _DEBUG
  int didwarn = 0;
#endif
  
  ex->stackmax = 0;
  for ( node = ex->start+1 ; node->type != FINISHED ; node++ )
  { usage += node->stack_delta;
    if ( usage > ex->stackmax ) ex->stackmax = usage;
#ifdef _DEBUG
    if ( (usage < 0) && (!didwarn) ) 
    { sprintf(errmsg,"Internal error: stack usage negative, node type %d.\n",
        node->type);
      sprintf(errmsg+strlen(errmsg),"(source file %s, line %d)\n",
         file_names[node->file_no],node->line_no);
      kb_error(2670,errmsg,WARNING);
      didwarn = 1;
    }     
    node->stack_spot = usage;
#endif
  }
}

/***********************************************************************
*
* Function: mark_element_loops()
*
* Purpose: Mark flags of expression nodes that are within 
*          element loops, so can know when freeing discards is
*          ill-advised.   Recursively transverses tree pre-order.
*/

void mark_element_loops(root,mark)
struct treenode *root;
int mark;  /* whether root in loop */
{ if ( mark ) 
    root->flags |= IN_ELEMENT_LOOP;
  else mark = root->flags & IN_ELEMENT_LOOP;

  if ( root->left )
    mark_element_loops(root+root->left,mark);
  if ( root->right )
    mark_element_loops(root+root->right,mark);
}
